require 'sketchup.rb'

#Version 0.2 Pro
#Added OBJ UV bridge to external UV mapping software

#Version 0.1

#initial release

if( not file_loaded?("UVtools.rb") )

menu = UI.menu("Plugins").add_submenu("UV Tools")
menu.add_item("Export UVs to OBJ")  { UVtools.start_bridge()}
menu.add_item("Import UVs from OBJ ") { UVtools.return_bridge() }

UI.add_context_menu_handler {|menu|
	ents=Sketchup.active_model.selection
	sel=Sketchup.active_model.selection.first
	sub=menu.add_submenu("UV Tools")
	sub.add_item("Spherical Map") {(UVtools.new.spherical_map(ents))}
	sub.add_item("Cylindrical Map") {(UVtools.new.cylindrical_map(ents))}
	
	
}

end
file_loaded("UVtools.rb")

class UVtools

@@selected_faces=[]

def initialize
	@points=[]
	@box=nil
end


######################
def get_bounds(ents)


all_points=[]
ents.each {|e|
	if e.class==Sketchup::Face
		all_points.push(e.vertices)
	end
}

return if all_points.empty?

all_points.flatten!
all_points.uniq!

box=Geom::BoundingBox.new
all_points.each {|v| box.add(v.position)}

@box=box

end

######################
def get_center(ents)  #returns the bounding box center of the selection

get_bounds(ents)
return @box.center

end

######################
def get_zmax_min(ents)  

get_bounds(ents)
return [@box.max.z,@box.min.z]

end
########################
def get_spherical_uv(p,c)  #p is the point we want to get the uvs for, c is the spherical center

pi=Math::PI
dx=p.x-c.x
dy=p.y-c.y
dz=p.z-c.z

theta=Math.acos(dz/(Math.sqrt(dx**2+dy**2+dz**2)))
phi=Math.atan2(dy,dx)

v=1.0-(theta/pi)

if phi<0.0 
	phi=2*pi+phi
end

u=(phi/(2*pi))

return [u,v,0.0]

end

########################
def get_cylindrical_uv(p,z_extents)  #p is the point we want to get the uvs for, ymin is the base of the cylinder

c=Geom::Point3d.new(@box.center)

pi=Math::PI
dx=p.x-c.x
dy=p.y-c.y

zmax=z_extents[0]
zmin=z_extents[1]
dz=(p.z-zmin)/(zmax-zmin)


phi=Math.atan2(dy,dx)

v=dz

if phi<0.0 
	phi=2*pi+phi
	#phi=-phi
end

u=phi/(2*pi)

return [u,v,0.0]

end

##########################
def validate_uvs(uvs)

u_arr=uvs.collect {|p| p.x}
u_max=u_arr.max
u_arr.each_index {|i|
	if u_arr[i]!=u_max
		if (u_max-u_arr[i])>0.8  #check for overlap
			#p "overlapped"
			u_arr[i]+=1.0
		end
	end
}

uvs.each_index {|i| uvs[i].x=u_arr[i]}  #modify the original uv array to fix the overlap

return uvs

end

##############################
def resize_uvs(uvs,mat)

tex=mat.texture
return if tex==nil

h=tex.height
w=tex.width

new_uvs=[]
uvs.each_index {|i|
	new_uvs[i]=Geom::Point3d.new(0,0,0)
	new_uvs[i].x=uvs[i].x*w
	new_uvs[i].y=uvs[i].y*h
	new_uvs[i].z=uvs[i].z
}

return new_uvs

end

 
#########################
def spherical_map(faces)  #performs spherical mapping

Sketchup.active_model.start_operation("Spherical Map")
faces=get_faces(faces)
origin=get_center(faces)
p origin

faces.each {|f|
	material=f.material
	verts=f.vertices
	pts=verts.collect {|v| v.position}
	uvs=pts.collect {|p| get_spherical_uv(p,origin)}
	uvs=validate_uvs(uvs)    #fixes the 'overlap' problem
	#uvs=resize_uvs(uvs,material)  #resize the uvs based on texture size
	if pts.length==3
		pt_array=[pts[0],uvs[0],pts[1],uvs[1],pts[2],uvs[2]]
	else
		pt_array=[pts[0],uvs[0],pts[1],uvs[1],pts[2],uvs[2],pts[3],uvs[3]]
	end
	begin
		f.position_material(material,pt_array,true)
		#f.position_material(material,pt_array,false)
	rescue
		p pt_array
		raise
	end
	
}

Sketchup.active_model.commit_operation
end	

#########################
def cylindrical_map(faces)  #performs spherical mapping

Sketchup.active_model.start_operation("Cylindrical Map")
faces=get_faces(faces)
z_extents=get_zmax_min(faces)

faces.each {|f|
	material=f.material
	verts=f.vertices
	pts=verts.collect {|v| v.position}
	uvs=pts.collect {|p| get_cylindrical_uv(p,z_extents)}
	uvs=validate_uvs(uvs)    #fixes the 'overlap' problem
	#uvs=resize_uvs(uvs,material)  #resize the uvs based on texture size
	if pts.length==3
		pt_array=[pts[0],uvs[0],pts[1],uvs[1],pts[2],uvs[2]]
	else
		pt_array=[pts[0],uvs[0],pts[1],uvs[1],pts[2],uvs[2],pts[3],uvs[3]]
	end
	begin
		f.position_material(material,pt_array,true)
		#f.position_material(material,pt_array,false)
	rescue
		p pt_array
		raise
	end
	
}

Sketchup.active_model.commit_operation
end	

############################
def get_faces(ents)

return ents.find_all {|e| e.class==Sketchup::Face}

end


######new in UVTools Pro
#collect the selected faces and export to OBJ
	def self.start_bridge()
	
	@@selected_faces=[]
	begin
		sel=Sketchup.active_model.selection
		sel_faces=self.get_faces(sel)
		@@selected_faces=sel_faces
		@materials=sel_faces.collect {|f| f.material}
		@materials.uniq!
		return if sel.length==nil
		
		obj_path=self.get_obj_path
		#path=UI.savepanel("Export UVs to OBJ",obj_path,"uv_in.obj")
		#return if not path
			
		mtl_path=File.join(File.dirname(obj_path),"su_uv_export.mtl")
		
		self.export_mtl(mtl_path,@materials)
		self.export_obj_faces(obj_path,sel_faces)
		
	rescue
		UI.messagebox("Unable to export UVs.")
		raise
	end
	
	UI.messagebox("OBJ exported to #{obj_path}.  Open this file with an external UV mapping aplication and then save it using the same path and filename when you have finished UV mapping.  Then, use the import command.")
	
	end
	
	#################
	def self.return_bridge()
	
	begin
		#sel=Sketchup.active_model.selection
		#sel_faces=self.get_faces(sel)
		sel_faces=@@selected_faces
		return if sel_faces.length==nil
		
		obj_path=self.get_obj_path
		num_failed=self.obj_uv_imp(sel_faces,obj_path)
	rescue
		UI.messagebox("Unable to import UVs from OBJ.")
		raise
	end
	UI.messagebox("UV import successful. #{num_failed} faces failed.")
	
	end

	######################
	def self.export_obj_faces(path,faces)
	
	@face_mesh=self.build_mesh(faces)
	@tw=Sketchup.create_texture_writer
	@v_index=0
	@uv_index=0
	@f_index=0
	out=File.new(path,"w")
	out.puts "# Alias OBJ Model file"
	out.puts "# Exported using Whaat's UV Tools"
	out.puts ""
	out.puts "mtllib su_uv_export.mtl"
	out.puts ""
	out.puts "g selected_faces"
	out.puts ""
	out.puts "vn 0 0 1"
	#self.export_vert_normals(out)
	self.export_verts(out)
	faces.each {|f|
		self.export_face(out,f)
	}
	
	out.close
	@tw=nil
	
	end
	
	####create mtl file
	def self.export_mtl(path,materials)
	
	#p path
	obj_path=self.get_obj_path
	out=File.new(path,"w")
	out.puts "#"
	out.puts "# Alias OBJ Material File"
	out.puts "# Exported using Whaat's UV Tools"
	out.puts ""
	@materials.each {|mat|
		if mat!=nil
			color=mat.color
			r=color.red.to_f/255.0
			g=color.green.to_f/255.0
			b=color.blue.to_f/255.0
			out.puts "newmtl #{mat.display_name}"
			out.puts "Ka 0.000000 0.000000 0.000000"
			out.puts "Kd #{r} #{g} #{b}"
			out.puts "Ks 0.330000 0.330000 0.330000"
			if mat.texture!=nil
				f_name=File.basename(mat.texture.filename)
				tex_path=File.join(File.dirname(obj_path),f_name)
				self.save_mat_texture(mat,tex_path)
				out.puts "map_Kd #{f_name}"
			end
		else
			r=0.8
			g=0.8
			b=0.8
			out.puts "newmtl Default"
			out.puts "Ka 0.000000 0.000000 0.000000"
			out.puts "Kd #{r} #{g} #{b}"
			out.puts "Ks 0.330000 0.330000 0.330000"
		end
	}
	out.close()
	
	end
	
	
	#export all the vertices in the face mesh
	def self.export_verts(out)
	
	pt=Geom::Point3d.new(0,0,0)
	vec=Geom::Vector3d.new(1,0,0)
	angle=-90.0.degrees
	
	obj_trans=Geom::Transformation.rotation(pt,vec,angle)
	pts=@face_mesh.points
	pts=pts.collect {|p| p.transform(obj_trans)}  #rotate all points by 90 degrees
	pts.each {|p|
		out.puts "v #{p.x.to_f} #{p.y.to_f} #{p.z.to_f}}"
	}
	
	end
		
	#acquire all faces in an array of entities
	def self.get_faces(ents)
	
	faces=[]
	faces=ents.find_all {|e| e.class==Sketchup::Face}
	return faces
	
	end
	
	#create a mesh
	def self.build_mesh(faces)
	
	mesh=Geom::PolygonMesh.new
	for f in faces
		verts=f.outer_loop.vertices
		verts.each {|v|
			mesh.add_point(v.position)
		}
	end
	
	return mesh

	end
	
	#export vertices, uvs, and face
	def self.export_face(out,f)
	if f.material==nil
		mat_name="Default"
	else
		mat_name=f.material.display_name
	end
	
	out.puts "usemtl #{mat_name}"
	mesh=f.mesh 5
	verts=f.outer_loop.vertices
	#v_indices=verts.collect {|v| @face_mesh.point_index(v.position)}
	uvh=f.get_UVHelper(true,false,@tw)
	face=["f"]  #array will hold the indices of the vertices and uvs
	#face=[]
	verts.each {|v|
		pos=v.position
		uv=uvh.get_front_UVQ(pos)
		#out.puts "v #{pos.x.to_f} #{pos.y.to_f} #{pos.z.to_f}"
		#@v_index+=1
		out.puts "vt #{uv.x.to_f} #{uv.y.to_f}"
		@uv_index+=1
		face.push("#{@face_mesh.point_index(pos)}\/#{@uv_index}\/1")
		#face.push("#{@face_mesh.point_index(pos)}\/#{@uv_index}")
	}
	#face.push("f")
	#face.reverse!
	out.puts face.join(" ")
	@f_index+=1
	out.puts ""
	
	end
	#export vertices, uvs, and face
	def self.export_face_void(out,f)
	
	mesh=f.mesh 5
	verts=f.outer_loop.vertices
	uvh=f.get_UVHelper(true,false,@tw)
	face=["f"]  #array will hold the indices of the vertices and uvs
	verts.each {|v|
		pos=v.position
		uv=uvh.get_front_UVQ(pos)
		out.puts "v #{pos.x.to_f} #{pos.y.to_f} #{pos.z.to_f}"
		@v_index+=1
		out.puts "vt #{uv.x.to_f} #{uv.y.to_f}"
		@uv_index+=1
		face.push("#{@v_index}\/#{@uv_index}\/1")
	}
	
	out.puts face.join(" ")
	@f_index+=1
	out.puts ""
	
	end
		
		
   def self.obj_uv_imp(faces,obj_path)
   version=Sketchup.version
	Sketchup.active_model.start_operation("Import OBJ uvs",true)  unless version<"7"
	Sketchup.active_model.start_operation("Import OBJ uvs")  if version<"7"
    	 failed=0
	@current_mat=nil
      @groups = {}
      @vertices = []
	   @uvs=[]
      @normals = []
      @faces = []
	  @mats=[]
	  face_index=0
      errors = Hash.new(0)
      face_verts = Hash.new(0)

      #f = UI.openpanel "Select .obj file", "", "*.obj"
      #return unless f
	  
	 f=obj_path
	  #f="C:\\Users\\Dale\\Desktop\\OBJ\\uv_bridge_out.obj"
	  obj_dir=File.dirname(f)
      line_cnt = g_cnt = 0
	  lines = IO.readlines(f)
	  num_lines=lines.length
      lines.each do |line|
	 line_cnt += 1
	 Sketchup.set_status_text("Processed #{line_cnt} of #{num_lines} lines")
	 next if line[0] == ?#
	    line.strip!
	 values = line.split
	 next if values.length == 0
	 cmd = values.shift

	 case cmd
	 when "mtllib"
		#mtl_path=File.join(obj_dir,values[0])
		#@mats=self.mtl_imp(mtl_path)  #import all the materials in the MTL file
	 when "usemtl"
		#mat_name=values[0]
		#@current_mat=@mats.find {|mat| mat.name==mat_name}  
	 when "v"
	   #v = values.map { |e| Float(e) * scale }
	    #@vertices << v
	   #mesh.add_point v if use_mesh
	when "vt"
		uvw=values.map{ |e| e.to_f }
		@uvs.push(uvw)

	 when "g"
	    #
	 when "f"
	    face = []
		face_index+=1
		face_uvs=[]
	    values.each do |v|
	       w = v.split("/")
	       face << Integer(w[0])
		    face_uvs.push(w[1].to_i)
	    end
	    #verts = face.map { |v| @vertices[v-1] }
		face_uvs=face_uvs.collect {|index| @uvs[index-1]}
	    #face_verts[verts.length] += 1
	    @faces << face
	   
	   begin
	      		current_face=faces[face_index-1]
				verts=current_face.outer_loop.vertices.collect {|v| v.position}
				#verts.reverse!
				l = verts.length
				current_mat=current_face.material
				if current_mat!=nil
					pt_array=[]
					pt_array=[verts[0],face_uvs[0],verts[1],face_uvs[1],verts[2],face_uvs[2]] if l==3
					pt_array=[verts[0],face_uvs[0],verts[1],face_uvs[1],verts[2],face_uvs[2],verts[3],face_uvs[3]] if l>3
					current_face.position_material(current_mat,pt_array,true)
				end
	       
	    rescue
	      #p "failed"
		  failed+=1
	    end
	 end
      end # case
    
	   Sketchup.active_model.commit_operation
	   return failed
   end
  
########################  
def self.save_mat_texture(mat,path)
  
Sketchup.active_model.start_operation("Save Texture")
tw=Sketchup.create_texture_writer
entities = Sketchup.active_model.active_entities
helper_group=entities.add_group
if (mat.respond_to?(:texture) and mat.texture !=nil)
	helper_group.material=mat
	tw.load(helper_group)
	status = tw.write(helper_group, path)
end	

#entities.erase_entities helper_group
Sketchup.active_model.abort_operation
tw=nil
if status==0   #write successful so apply the image to the dialog
	return path
else
	return nil
end

end  

###get the default path to write the OBJ to
def self.get_obj_path

obj_name="su_uv_bridge.obj"
on_mac=(Object::RUBY_PLATFORM =~ /mswin/i) ? false : ((Object::RUBY_PLATFORM =~ /darwin/i) ? true : :other)

if on_mac
	path=File.join(ENV['HOME'],"Library","Application Support","UVTools")
else
	path=File.join(ENV['APPDATA'],"UVTools")
end

Dir.mkdir(path) unless FileTest.exist?(path)

return File.join(path,obj_name)

end
  

end #class